"""
Created on Thu Sep 10 11:58:46 2015

This example demonstrates the (lack of) reliability of transforming between DFT
represented by real and imaginary parts and its representation by amplitude and phase.

Copyright © 2021 Sascha Eichstädt, Björn Ludwig (PTB)
"""
import matplotlib.pyplot as plt
import numpy as np


def demonstrate_GUM_dft2ampphase():
    rst = np.random.RandomState(1)

    uy = 0.001

    def GUM(y):
        x = np.sqrt(y[0] ** 2 + y[1] ** 2)
        ux = np.sqrt((y[0] / x * uy) ** 2 + (y[1] / x * uy) ** 2)
        return x, ux

    def MonteCarlo(y, runs=10000):
        yMC = np.tile(y, (runs, 1)) + rst.randn(runs, 2) * uy
        xMC = np.sqrt(yMC[:, 0] ** 2 + yMC[:, 1] ** 2)
        x = np.mean(xMC)
        ux = np.std(xMC)
        return x, ux

    N = 50
    y1 = np.logspace(np.log10(uy * 100), -6, N)
    y2 = np.logspace(np.log10(uy * 100), -6, N)

    xGUM = np.zeros(N)
    uxGUM = np.zeros(N)
    xMC = np.zeros(N)
    uxMC = np.zeros(N)

    for k in range(len(y1)):
        xGUM[k], uxGUM[k] = GUM(np.r_[y1[k], y2[k]])
        xMC[k], uxMC[k] = MonteCarlo(np.r_[y1[k], y2[k]])
        print(".", end=" ")
    print("")

    error = np.abs(xGUM - xMC)
    print("max error is %g" % error.max())
    u_error = np.abs(uxGUM - uxMC)
    print("max u_error is %g" % u_error.max())

    plt.figure(1)
    plt.clf()
    plt.semilogx(y1 / uy, xGUM, label="GUM")
    plt.semilogx(y1 / uy, xMC, label="Monte Carlo")
    plt.legend(loc="best")
    plt.xlabel(r"$y/uy$ in a.u.")
    plt.ylabel(r"estimate of $\sqrt{y_1^2+y_2^2}$")

    plt.figure(2)
    plt.clf()
    plt.semilogx(y1 / uy, uxGUM, label="GUM")
    plt.semilogx(y1 / uy, uxMC, label="Monte Carlo")
    plt.legend(loc="best")
    plt.xlabel(r"$y/uy$ in a.u.")
    plt.ylabel(r"$u(\sqrt{y_1^2+y_2^2})$")

    plt.figure(3)
    plt.clf()
    plt.semilogx(y1 / uy, error, label="error in estimate")
    plt.semilogx(y1 / uy, u_error, label="error in uncertainty")
    plt.legend(loc="best")
    plt.xlabel(r"$y/uy$ in a.u.")

    plt.show()


if __name__ == "__main__":
    demonstrate_GUM_dft2ampphase()
