"""
Created on Thu Sep 10 11:58:46 2015

In this example an actually measured signal is deconvolved and the result compared to a
reference measurement signal.

Copyright © 2021 Sascha Eichstädt, Björn Ludwig (PTB)
"""
import os

import matplotlib.pyplot as plt
import numpy as np

from GUM2DFT import AmpPhase2DFT, GUM_DFT, GUM_iDFT, GUMdeconv


def example_deconvolution_with_MonteCarlo():
    file_directory = os.path.dirname(os.path.abspath(__file__))

    # set up colors for plotting
    colors = [[0.1, 0.6, 0.5], [0.9, 0.2, 0.5], [0.9, 0.5, 0.1]]
    # seed random number generator
    rst = np.random.RandomState(1)
    # number of Monte Carlo runs for comparison
    MCruns = 10000

    # low-pass filter for deconvolution
    def lowpass(f, fcut=80e6):
        return 1 / (1 + 1j * f / fcut) ** 2

    # reference data
    ref_file = np.loadtxt(os.path.join(file_directory, "data", "reference_signal.dat"))
    time = ref_file[:, 0]
    ref_data = ref_file[:, 1]
    N = len(time)
    Ts = 2e-9

    # measured hydrophone output signal
    meas = np.loadtxt(os.path.join(file_directory, "data", "measured_signal.dat"))
    y = meas[:, 1]
    # assumed noise std
    noise_std = 4e-4
    Uy = noise_std ** 2

    # hydrophone calibration data
    calib = np.loadtxt(os.path.join(file_directory, "data", "calib_PA055.dat"))
    f = calib[:, 0]
    FR = calib[:, 1] * np.exp(1j * calib[:, 3])
    Nf = 2 * (len(f) - 1)

    uAmp = calib[:, 2]
    uPhas = calib[:, 4]

    # low pass filter for deconvolution regularization
    HL = lowpass(f)
    # inverse system for input estimation
    Hc = FR / np.abs(HL)
    uAmp /= np.abs(HL)
    H = np.r_[np.real(Hc), np.imag(Hc)]

    UAP = np.r_[uAmp, uPhas * np.pi / 180] ** 2
    n = len(UAP) // 2

    # estimation of measurand using quasi-static approach
    f_awf = 2.927 * 1e6
    M_awf = 0.19341
    uM_awf = M_awf * 0.05
    xh_awf = y / M_awf
    uxh_awf = np.sqrt(Uy / M_awf ** 2 + y ** 2 / M_awf ** 4 * uM_awf ** 2)

    def deconv_GUM2DFT():  # small uncertainties for low and high frequencies
        print("Propagating uncertainty associated with measurement through DFT")
        Y, UY = GUM_DFT(y, Uy, N=Nf)
        # propagation to real/imag
        print(
            "Propating uncertainty associated with calibration data to real and imag part"
        )
        UH = AmpPhase2DFT(np.abs(Hc), np.unwrap(np.angle(Hc)) * np.pi / 180, UAP)[1]
        # propagation through deconvolution equation
        print("Propating uncertainty through the deconvolution")
        XH, UXH = GUMdeconv(H, Y, UH, UY)
        # propagation back to time domain
        print(
            "Propagating uncertainty associated with the estimate back to time domain\n"
        )
        xh, Uxh = GUM_iDFT(XH, UXH, Nx=N)
        return xh, Uxh

    def deconv_MonteCarlo():
        print("Carrying out Monte Carlo with %d runs" % MCruns)
        RIMC = np.zeros((MCruns, 2 * len(FR)))
        for k in range(MCruns):
            A = np.abs(Hc) + rst.randn(n) * uAmp
            P = np.unwrap(np.angle(Hc)) + rst.randn(n) * uPhas
            P *= np.pi / 180
            RIMC[k, :] = np.r_[A * np.cos(P), A * np.sin(P)]

        yMC = np.tile(y, (MCruns, 1)) + rst.randn(MCruns, len(y)) * np.sqrt(Uy)
        YMC = np.fft.rfft(yMC, n=Nf)
        FMC = RIMC[:, :n] + 1j * RIMC[:, n:]
        XMC = YMC / FMC

        xMC = np.fft.irfft(XMC)[:, :N]
        xh = np.mean(xMC, axis=0)
        Uxh = np.cov(xMC, rowvar=0)
        return xh, Uxh

    """Carrying out the calculations"""
    xh, Uxh = deconv_GUM2DFT()
    xhMC, UxhMC = deconv_MonteCarlo()

    ux = np.sqrt(np.diag(Uxh))
    uxMC = np.sqrt(np.diag(UxhMC))

    # calculation of peak values
    Nt = np.nonzero((time <= 2e-6))[0][-1]
    u_ref = 0.2
    max_peaks = np.array(
        [ref_data[:Nt].max(), xh[:Nt].max(), xhMC[:Nt].max(), xh_awf[:Nt].max()]
    )
    min_peaks = np.array(
        [ref_data[:Nt].min(), xh[:Nt].min(), xhMC[:Nt].min(), xh_awf[:Nt].min()]
    )
    unc_max = np.array(
        [
            u_ref,
            ux[xh[:Nt].argmax()],
            uxMC[xhMC[:Nt].argmax()],
            uxh_awf[xh_awf[:Nt].argmax()],
        ]
    )
    unc_min = np.array(
        [
            u_ref,
            ux[xh[:Nt].argmin()],
            uxMC[xhMC[:Nt].argmin()],
            uxh_awf[xh_awf[:Nt].argmin()],
        ]
    )

    #%% plotting results

    plt.figure(1)
    plt.clf()
    plt.plot(time * 1e6, xh, label="estimate (GUM2DFT)", linewidth=2, color=colors[0])
    plt.plot(
        time * 1e6,
        xh_awf,
        "-.",
        label="estimate (quasi-static)",
        linewidth=2,
        color=colors[2],
    )
    plt.plot(
        time * 1e6,
        ref_data,
        "--",
        label="reference data",
        linewidth=2,
        color=colors[1],
    )
    plt.fill_between(time * 1e6, xh + 2 * ux, xh - 2 * ux, alpha=0.2, color=colors[0])
    plt.fill_between(
        time * 1e6,
        xh_awf + 2 * uxh_awf,
        xh_awf - 2 * uxh_awf,
        alpha=0.2,
        color=colors[2],
    )
    plt.xlabel(r"time / $\mu$s", fontsize=22)
    plt.ylabel("signal amplitude / MPa", fontsize=22)
    plt.tick_params(which="major", labelsize=18)
    plt.legend(loc="upper left", fontsize=18, fancybox=True)
    plt.xlim(0, 2)

    dB = lambda a: 20 * np.log10(np.abs(a))
    plt.figure(2)
    plt.clf()
    plt.plot(f * 1e-6, dB(FR), label="hydrophone", linewidth=2)
    plt.plot(f * 1e-6, dB(1 / Hc), label="deconvolution", linewidth=2)
    plt.plot(f * 1e-6, dB(FR / Hc), "--", linewidth=0.5)
    plt.legend(loc="best", fontsize=18)
    plt.xlim(0.5, 100)
    plt.ylim(-40, 40)
    plt.xlabel("frequency / MHz", fontsize=22)
    plt.ylabel("frequency response amplitude / dB", fontsize=22)
    plt.tick_params(which="both", labelsize=18)

    plt.figure(3)
    plt.clf()
    plt.subplot(211)
    plt.errorbar(
        f * 1e-6,
        np.abs(Hc),
        2 * np.sqrt(UAP[: len(UAP) // 2]),
        fmt=".-",
        alpha=0.2,
        color=colors[0],
    )
    plt.xlim(0.5, 80)
    plt.ylim(0.04, 0.24)
    plt.xlabel("frequency / MHz", fontsize=22)
    plt.tick_params(which="both", labelsize=18)
    plt.ylabel("amplitude / V/MPa", fontsize=22)
    plt.subplot(212)
    plt.errorbar(
        f * 1e-6,
        np.unwrap(np.angle(Hc)) * np.pi / 180,
        2 * UAP[len(UAP) // 2 :],
        fmt=".-",
        alpha=0.2,
        color=colors[0],
    )
    plt.xlim(0.5, 80)
    plt.ylim(-0.2, 0.3)
    plt.xlabel("frequency / MHz", fontsize=22)
    plt.tick_params(which="both", labelsize=18)
    plt.ylabel("phase / rad", fontsize=22)

    plt.figure(4)
    plt.clf()
    plt.plot(time * 1e6, ux, label="GUM2DFT", linewidth=2, color=colors[0])
    plt.plot(time * 1e6, uxMC, "k--", label="Monte Carlo", linewidth=2)
    plt.plot(time * 1e6, uxh_awf, label="quasi-static", linewidth=2, color=colors[2])
    plt.xlabel(r"time / $\mu$s", fontsize=22)
    plt.ylabel("uncertainty / MPa", fontsize=22)
    plt.tick_params(which="major", labelsize=18)
    plt.legend(loc="best", fontsize=19)
    plt.xlim(0, 2)

    print("\npositive pressure peak")
    print("reference  | GUM2DFT | Monte Carlo | awf |")
    print(
        "%.2g (%.2g) | %.2g (%.2g) | %.2g (%.2g) | %.2g (%.2g)"
        % tuple([m for sm in zip(max_peaks, unc_max) for m in sm])
    )

    print("\nnegative pressure peak")
    print("reference  | GUM2DFT | Monte Carlo | awf")
    print(
        "%.2g (%.2g) | %.2g (%.2g) | %.2g (%.2g) | %.2g (%.2g)"
        % tuple([m for sm in zip(-min_peaks, unc_min) for m in sm])
    )

    plt.show()


if __name__ == "__main__":
    example_deconvolution_with_MonteCarlo()
