"""
Created on Thu Sep 10 11:58:46 2015

In this example an actually measured signal is deconvolved and the result compared to a
reference measurement signal.

Copyright © 2021 Sascha Eichstädt, Björn Ludwig (PTB)
"""

import os

import matplotlib.pyplot as plt
import numpy as np

from GUM2DFT import GUM_DFT, GUM_iDFT, GUM_multiply


def demonstrate_multiplication():
    """Conduct the actual multiplication and plot results."""
    file_directory = os.path.dirname(os.path.abspath(__file__))

    # low-pass filter
    def lowpass(f, fcut=100e6):
        return 1 / (1 + 1j * f / fcut) ** 2

    # measured hydrophone output signal
    meas = np.loadtxt(os.path.join(file_directory, "data", "measured_signal.dat"))
    time = meas[:, 0]
    y = meas[:, 1]
    Ts = 2e-9
    # assumed noise std
    noise_std = 1e-3
    Uy = np.ones(len(y)) * noise_std ** 2

    # %% filter for multiplication
    Nf = 4096
    f = np.fft.rfftfreq(Nf, Ts)
    HL = lowpass(f)
    HL = np.r_[np.real(HL), np.imag(HL)]

    print("Propagating uncertainty associated with measurement through DFT")
    Y, UY = GUM_DFT(y, Uy, N=Nf)
    print("Propagating uncertainty through the multiplication in the DFT domain")
    YL, UYL = GUM_multiply(Y=Y, F=HL, UY=UY)  # type: np.ndarray
    print("Propagating uncertainty through inverse DFT")
    yL, UyL = GUM_iDFT(YL, UYL, Nx=len(y))

    uy = np.sqrt(Uy)
    uyL = np.sqrt(np.diag(UyL))

    # plotting results

    plt.figure(1)
    plt.clf()
    plt.plot(time * 1e6, y, label="original signal")
    plt.plot(time * 1e6, yL, label="filtered signal")
    plt.legend(loc="best")
    plt.show()

    plt.figure(2)
    plt.clf()
    plt.plot(
        np.sqrt(np.diag(UY[: Nf // 2, : Nf // 2])),
        "+",
        label="original signal",
        alpha=0.5,
    )
    plt.plot(
        np.sqrt(np.diag(UYL[: Nf // 2, : Nf // 2])),
        "o",
        label="filtered " "signal",
        alpha=0.2,
    )
    plt.legend(loc="best")
    plt.show()

    plt.figure(3)
    plt.clf()
    plt.plot(time * 1e6, uy, "+", label="original signal", alpha=0.5)
    plt.plot(time * 1e6, uyL, "o", label="filtered signal", alpha=0.2)
    plt.legend(loc="best")
    plt.ylim(0, 2 * noise_std)
    plt.show()


if __name__ == "__main__":
    demonstrate_multiplication_example()
