"""
Created on Fri Aug 21 11:22:21 2015

This is a clean simulated deconvolution example. That is, using the measured frequency
response an output signal is generated by means of convolution and thereafter the
input signal estimated by deconvolution. The advantage of this approach is to verify
the general applicability of the method.

Copyright © 2021 Sascha Eichstädt, Björn Ludwig (PTB)
"""
import os

from matplotlib.pyplot import *
from statsmodels.tsa.arima_process import ArmaProcess

from examples.tools import multivariate_MonteCarlo as MC
from GUM2DFT import AmpPhase2DFT, GUM_DFT, GUM_iDFT, GUMdeconv


def demonstrate_simulation():
    """Conduct the actual simulation and plot results."""
    file_directory = os.path.dirname(os.path.abspath(__file__))
    # low-pass filter for deconvolution

    def lowpass(f, fcut=120e6):
        return 1 / (1 + 1j * f / fcut) ** 2

    # definition of input data
    input_file = os.path.join(file_directory, "data", "reference_signal.dat")
    noise_std = 1e-4

    time, pressure = np.loadtxt(input_file).T
    Ts = 2e-9

    # simulation of measurement w/o noise
    calib_file = os.path.join(file_directory, "data", "calib_PA055.dat")
    calib = np.loadtxt(calib_file)
    f = calib[:, 0]
    Nf = 2 * len(f) - 1
    FR = calib[:, 1] * np.exp(1j * calib[:, 3])
    P = np.fft.rfft(np.r_[pressure, np.zeros(2 * len(f) - len(pressure) - 1)])
    x = np.fft.irfft(FR * P)[: len(time)]

    uAmp = calib[:, 2]
    uPhas = calib[:, 4]

    # adding correlated noise (ARMA process)
    N = len(time)
    ar = np.array([0.75, -0.6])
    ma = np.array([0.45, 0.1])
    ARMA = ArmaProcess(ar, ma)
    noise = ARMA.generate_sample(N) * 0.01

    x += np.random.randn(len(time)) * noise_std

    acov = ARMA.acovf(nobs=N) * noise_std ** 2

    Ux = np.zeros((N, N))
    for k in range(N):
        Ux[k, k:] = acov[: N - k]
    Ux = Ux + Ux.T - np.diag(np.diag(Ux))

    # low pass filter for regularized deconvolution
    HL = lowpass(f)
    uAmp /= np.abs(HL)

    # propagation to real/imag domain
    UH = AmpPhase2DFT(
        np.abs(FR), np.angle(FR) * np.pi / 180, np.r_[uAmp, uPhas * np.pi / 180]
    )[1]
    # comparison to Monte Carlo propagation
    MCruns = 100000
    Nv = len(FR)
    MCfun = lambda val: np.r_[
        np.real(val[:Nv] * np.exp(1j * val[Nv:])),
        np.imag(val[:Nv] * np.exp(1j * val[Nv:])),
    ]
    mean = np.r_[np.abs(FR), np.angle(FR)]
    covariance = np.r_[uAmp, uPhas * np.pi / 180]
    HRI, RI, UHMC = MC(mean, covariance, MCfun, MCruns)

    # inverse system for input estimation
    Hc = FR / HL
    H = np.r_[np.real(Hc), np.imag(Hc)]

    # propagation from time to frequency domain
    Y, UY = GUM_DFT(x, Ux, N=Nf)

    # propagation through deconvolution equation
    XH, UXH = GUMdeconv(H, Y, UH, UY)

    # propagation back to time domain
    xh, Uxh = GUM_iDFT(XH, UXH, Nx=len(x))

    # plotting results

    figure(1)
    clf()
    plot(time * 1e6, pressure, label="input pressure", linewidth=2)
    plot(time * 1e6, x, label="measurement", linewidth=2)
    xlabel(r"time / $\mu$s", fontsize=22)
    ylabel("signal amplitude / a.u.", fontsize=22)
    tick_params(which="major", labelsize=18)
    legend(loc="best", fontsize=20, fancybox=True)
    xlim(0.5, 1.5)
    figure(1).show()

    ux = np.sqrt(np.diag(Uxh))

    figure(2)
    clf()
    plot(time * 1e6, xh, label="estimate of measurand", linewidth=2)
    plot(time * 1e6, pressure, "--", label="true value of measurand", linewidth=2)
    fill_between(time * 1e6, xh + ux, xh - ux, alpha=0.2)
    xlabel(r"time / $\mu$s", fontsize=22)
    ylabel("signal amplitude / a.u.", fontsize=22)
    tick_params(which="major", labelsize=18)
    legend(loc="best", fontsize=20, fancybox=True)
    xlim(0.5, 1.5)
    figure(2).show()

    figure(3)
    clf()
    semilogy(f * 1e-6, np.abs(FR), label="hydrophone", linewidth=2)
    semilogy(f * 1e-6, np.abs(1 / Hc), label="deconvolution", linewidth=2)
    semilogy(f * 1e-6, np.abs(FR / Hc), "--", linewidth=0.5)
    legend(loc="best", fontsize=18)
    xlim(0.5, 150)
    xlabel("frequency / MHz", fontsize=22)
    ylabel("frequency resp. ampl. / a.u.", fontsize=22)
    tick_params(which="both", labelsize=18)
    figure(3).show()

    figure(4)
    clf()
    subplot(211)
    errorbar(f * 1e-6, calib[:, 1], 2 * uAmp, fmt=".", alpha=0.2)
    xlim(0.5, 150)
    ylim(0, 0.25)
    xlabel("frequency / MHz", fontsize=22)
    tick_params(which="both", labelsize=18)
    ylabel("amplitude / a.u.", fontsize=22)
    subplot(212)
    errorbar(f * 1e-6, calib[:, 2], 2 * uPhas, fmt=".", alpha=0.2)
    xlim(0.5, 150)
    xlabel("frequency / MHz", fontsize=22)
    tick_params(which="both", labelsize=18)
    ylabel("phase / deg", fontsize=22)
    figure(4).show()

    figure(5)
    clf()
    subplot(211)
    plot(f * 1e-6, np.sqrt(np.diag(UH)[:Nv]), label="GUM formula")
    plot(f * 1e-6, np.sqrt(np.diag(UHMC)[:Nv]), label="Monte Carlo")
    title("Real part")
    legend(loc="best")
    subplot(212)
    plot(f * 1e-6, np.sqrt(np.diag(UH)[Nv:]), label="GUM formula")
    plot(f * 1e-6, np.sqrt(np.diag(UHMC)[Nv:]), label="Monte Carlo")
    legend(loc="best")
    title("Imaginary part")
    figure(5).show()


if __name__ == "__main__":
    demonstrate_simulation()
