"""
Created on Thu Sep 10 11:58:46 2015

Comparison of the effect of different window functions. Note that the warning issued by
GUM2DFT function DFT2AmpPhase is ignored here, because only a comparison of the
different windows is intended, for which the achieved precision appears to be enough.

Copyright © 2021 Sascha Eichstädt, Björn Ludwig (PTB)
"""

import matplotlib.pyplot as plt
import numpy as np
from scipy.signal import windows

from examples.tools import tukeywin
from GUM2DFT import apply_window, DFT2AmpPhase, GUM_DFT


def demonstrate_windowing():
    """Conduct the actual example calculation and plot results."""
    Ts = 2e-9
    time = np.arange(0, 8e-6, Ts)
    N = len(time)

    #%% original signal
    freqs = [2e6, 5e6, 10e6, 12e6]
    amps = [0.5, 2.2, 2.5, 1.8]
    y = np.array([a * np.sin(2 * np.pi * fr * time) for a, fr in zip(amps, freqs)]).sum(
        axis=0
    )
    Uy = 1e-2

    #%% w/o window
    rect = np.ones_like(time)
    Y, UY = GUM_DFT(y, Uy, window=rect)
    A, P, UAP = DFT2AmpPhase(Y, UY, keep_sparse=True, return_type="separate")
    fY = np.arange(0, N // 2 + 1) / (N * Ts)

    #%% with Tukey window
    tukey = tukeywin(len(y), alpha=0.5)
    yw, Uyw = apply_window(y, Uy, tukey)
    Yw, UYw = GUM_DFT(y, Uy, window=tukey)
    Aw, Pw, UAPw = DFT2AmpPhase(Yw, UYw, keep_sparse=True, return_type="separate")

    #%% with Hann window
    hann = windows.hann(len(y))
    yw2, Uyw2 = apply_window(y, Uy, hann)
    Yw2, UYw2 = GUM_DFT(y, Uy, window=hann)
    Aw2, Pw2, UAPw2 = DFT2AmpPhase(Yw2, UYw2, keep_sparse=True, return_type="separate")

    #%%
    plt.figure(1)
    plt.clf()
    plt.plot(time * 1e6, y, label="rectangular window")
    plt.plot(time * 1e6, yw, label="Tukey window")
    plt.plot(time * 1e6, yw2, label="Hann window")
    plt.xlabel(r"time / $\mu$s", fontsize=20)
    plt.ylabel("amplitude / au", fontsize=20)
    plt.legend(loc="best")

    plt.figure(2)
    plt.clf()
    plt.plot(time * 1e6, rect * np.sqrt(Uy), label="rectangular window")
    plt.plot(time * 1e6, np.sqrt(Uyw), label="Tukey window")
    plt.plot(time * 1e6, np.sqrt(Uyw2), label="Hann window")
    plt.legend(loc="best")
    plt.ylim(0, 1.2 * np.sqrt(Uy))
    plt.xlabel(r"time / $\mu$s", fontsize=20)
    plt.ylabel("window amplitude / au", fontsize=20)

    uAmp = np.sqrt(UAP[0][: N // 2 + 1]) / np.sum(rect)
    uAmpw = np.sqrt(UAPw[0][: N // 2 + 1]) / np.sum(tukey)
    uAmpw2 = np.sqrt(UAPw2[0][: N // 2 + 1]) / np.sum(hann)

    plt.figure(3)
    plt.clf()
    plt.errorbar(
        fY[: N // 2 + 1] * 1e-6,
        A / np.sum(rect),
        2 * uAmp,
        fmt=".-",
        label="rectangular window",
    )
    plt.errorbar(
        fY[: N // 2 + 1] * 1e-6,
        Aw / np.sum(tukey),
        2 * uAmpw,
        fmt=".-",
        label="Tukey window",
    )
    plt.errorbar(
        fY[: N // 2 + 1] * 1e-6,
        Aw2 / np.sum(hann),
        2 * uAmpw2,
        fmt=".-",
        label="Hann window",
    )
    plt.xlim(0, 15)
    plt.legend(loc="best")
    for fr in freqs:
        plt.axvline(fr * 1e-6, linestyle="--", color="k")
    plt.xlabel(r"frequency / MHz", fontsize=20)
    plt.ylabel("DFT amplitude / au", fontsize=20)

    print("\nDifference of endpoints w/o window \t", end=" ")
    print("%g" % (np.abs(y[0] - y[-1])))

    print("Difference of endpoints with Tukey window \t", end=" ")
    print("%g" % (np.abs(yw[0] - yw[-1])))

    plt.show()


if __name__ == "__main__":
    demonstrate_windowing()
