"""
The tools module is a collection of miscellaneous helper functions. It contains the
following functions:

- make_semiposdef(matrix, maxiter, tol, verbose):
      Make quadratic matrix positive semi-definite by increasing its eigenvalues
- multivariate_MonteCarlo(mean, covariance, modelfun, runs):
      Carry out multivariate Monte Carlo propagation
- tukeywin(window_length, alpha):
      Compute a Tukey window, also known as the cosine-tapered window

Copyright © 2021 Sascha Eichstädt, Björn Ludwig (PTB)

License
=======

This software is licensed under the LGPLv3 license.

When publishing results produced by this software, cite
S. Eichstädt, V. Wilkens
'GUM2DFT - A software tool for uncertainty evaluation of transient signals in the
frequency domain',
submitted to Meas. Sci. Technol.

DISCLAIMER
==========
This software was developed at Physikalisch-Technische Bundesanstalt
(PTB). The software is made available "as is" free of cost. PTB assumes
no responsibility whatsoever for its use by other parties, and makes no
guarantees, expressed or implied, about its quality, reliability, safety,
suitability or any other characteristic. In no event will PTB be liable
for any direct, indirect or consequential damage arising in connection
with the use of this software.
"""
from typing import Optional

import numpy as np
from scipy.sparse import eye, issparse
from scipy.sparse.linalg.eigen.arpack import eigs

__all__ = ["make_semiposdef", "multivariate_MonteCarlo", "tukeywin"]


def make_semiposdef(
    matrix: np.ndarray,
    maxiter: Optional[int] = 10,
    tol: Optional[float] = 1e-12,
    verbose: Optional[bool] = False,
) -> np.ndarray:
    """Make quadratic matrix positive semi-definite by increasing its eigenvalues

    Parameters
    ----------
    matrix : array_like of shape (N,N)
        the matrix to process
    maxiter : int, optional
        the maximum number of iterations for increasing the eigenvalues, defaults to 10
    tol : float, optional
        tolerance for deciding if pos. semi-def., defaults to 1e-12
    verbose : bool, optional
        If True print smallest eigenvalue of the resulting matrix, if False (default)
        be quiet

    Returns
    -------
    (N,N) array_like
        quadratic positive semi-definite matrix

    Raises
    ------
    ValueError
        If matrix is not square
    """
    n, m = matrix.shape
    if n != m:
        raise ValueError("Matrix has to be quadratic")
    # use specialised functions for sparse matrices
    if issparse(matrix):
        # enforce symmetric matrix
        matrix = 0.5 * (matrix + matrix.T)
        # calculate smallest eigenvalue
        e = np.real(eigs(matrix, which="SR", return_eigenvectors=False)).min()
        count = 0
        # increase the eigenvalues until matrix is positive semi-definite
        while e < tol and count < maxiter:
            matrix += (np.absolute(e) + tol) * eye(n, format=matrix.format)
            e = np.real(eigs(matrix, which="SR", return_eigenvectors=False)).min()
            count += 1
        e = np.real(eigs(matrix, which="SR", return_eigenvectors=False)).min()
    # same procedure for non-sparse matrices
    else:
        matrix = 0.5 * (matrix + matrix.T)
        count = 0
        e = np.real(np.linalg.eigvals(matrix)).min()
        while e < tol and count < maxiter:
            e = np.real(np.linalg.eigvals(matrix)).min()
            matrix += (np.absolute(e) + tol) * np.eye(n)
        e = np.real(np.linalg.eigvals(matrix)).min()
    if verbose:
        print("Final result of make_semiposdef: smallest eigenvalue is %e" % e)
    return matrix


def multivariate_MonteCarlo(mean, covariance, modelfun, runs):
    """Carry out multivariate Monte Carlo propagation

    Parameters
    ----------
    mean : np.ndarray
        vector of best estimate
    covariance : np.ndarray
        matrix of uncertainties either as covariance matrix or as vector of
        uncertainties (i.e. zero correlation)
    modelfun : callable
        function which takes arrays of dimension of mean and returns model output
    runs : int
        number of Monte Carlo trials

    Returns
    -------
    model_outputs : np.ndarray
        vector of model outputs, mean and covariance
    mean_estimate : np.ndarray
        MonteCarlo estimate of the model mean
    covariance_estimate : np.ndarray
        MonteCarlo estimate of the associated uncertainties

    Raises
    ------
    ValueError
        If model function 'modelfun' does not work with the inputs
    """
    try:
        test = modelfun(mean)
        output_dim = len(test)
    except:
        raise ValueError("Model function appears to be wrong.")

    if len(covariance.shape) == 1:  # zero correlation
        MCinputs = np.tile(mean, (runs, 1)) + np.random.randn(
            runs, len(mean)
        ) * np.tile(covariance, (runs, 1))
        model_outputs = np.zeros((runs, output_dim))
        for k in range(runs):
            model_outputs[k, :] = modelfun(MCinputs[k, :])
    else:
        model_outputs = np.random.multivariate_normal(mean, covariance, runs)

    estimate = model_outputs.mean(axis=0)
    uncertainty = np.cov(model_outputs, rowvar=False)
    return model_outputs, estimate, uncertainty


def tukeywin(window_length, alpha=0.5) -> np.ndarray:
    r"""Compute a Tukey window, also known as the cosine-tapered window

    The Tukey window can be regarded as a cosine lobe of width \alpha * N / 2 that is
    convolved with a rectangle window of width N * (1 - \alpha / 2).

    Downloaded from  http://leohart.wordpress.com/2006/01/29/hello-world/ at 2014-03-13

    Parameters
    ----------
    alpha : float
        Cosine fraction. At \alpha = 1 it becomes rectangular, and at \alpha = 0 it
        becomes a Hann window.
    window_length : int
        window length N

    Returns
    -------
    np.ndarray
        The desired Tukey window

    References
    ----------
    J.W. Tukey: An introduction to the calculations of numerical spectrum analysis.
    In: Spectral Analysis of Time Series. 1967, S. 25–46
    """

    # Special cases
    if alpha <= 0:
        return np.ones(window_length)  # rectangular window
    elif alpha >= 1:
        return np.hanning(window_length)

    # Normal case
    x = np.linspace(0, 1, window_length)
    w = np.ones(x.shape)

    # first condition 0 <= x < alpha/2
    first_condition = x < alpha / 2
    w[first_condition] = 0.5 * (
        1 + np.cos(2 * np.pi / alpha * (x[first_condition] - alpha / 2))
    )

    # second condition already taken care of

    # third condition 1 - alpha / 2 <= x <= 1
    third_condition = x >= (1 - alpha / 2)
    w[third_condition] = 0.5 * (
        1 + np.cos(2 * np.pi / alpha * (x[third_condition] - 1 + alpha / 2))
    )

    return w
